@@ -4,6 +4,13 @@
  * 	- from svgload.c
  * 25/4/16
  * 	- add giflib5 support
+ * 26/7/16
+ * 	- transparency was wrong if there was no EXTENSION_RECORD
+ * 	- write 1, 2, 3, or 4 bands depending on file contents
+ * 17/8/16
+ * 	- support unicode on win
+ * 19/8/16
+ * 	- better transparency detection, thanks diegocsandrim
  */
 
 /*
@@ -42,8 +49,6 @@
 #endif /*HAVE_CONFIG_H*/
 #include <vips/intl.h>
 
-#ifdef HAVE_GIFLIB
-
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -55,10 +60,12 @@
 #include <vips/internal.h>
 #include <vips/debug.h>
 
+#ifdef HAVE_GIFLIB
+
 #include <gif_lib.h>
 
 /* giflib 5 is rather different :-( functions have error returns and there's
- * not LastError function.
+ * no LastError().
  *
  * GIFLIB_MAJOR was introduced in 4.1.6. Use it to test for giflib 5.x.
  */
@@ -86,6 +93,18 @@ typedef struct _VipsForeignLoadGif {
 	 */
 	GifPixelType *line;
 
+	/* We decompress the whole thing to a huge RGBA memory image, and
+	 * as we render, watch for bands and transparency. At the end of
+	 * loading, we copy 1 or 3 bands, with or without transparency to
+	 * output.
+	 */
+	gboolean has_transparency;
+	gboolean has_colour;
+
+	/* The FILE* we read from.
+	 */
+	FILE *fp;
+
 } VipsForeignLoadGif;
 
 typedef VipsForeignLoadClass VipsForeignLoadGifClass;
@@ -200,24 +219,44 @@ vips_foreign_load_gif_close( VipsForeignLoadGif *gif )
 		gif->file = NULL;
 	}
 #endif
+
+	VIPS_FREEF( fclose, gif->fp ); 
+}
+
+/* Our input function for file open. We can't use DGifOpenFileName(), since
+ * that just calls open() and won't work with unicode on win32. We can't use
+ * DGifOpenFileHandle() since that's an fd from open() and you can't pass those
+ * acoss DLL boundaries on Windows. 
+ */
+static int 
+vips_giflib_file_read( GifFileType *file, GifByteType *buffer, int n )
+{
+	FILE *fp = (FILE *) file->UserData;
+
+	return( (int) fread( (void *) buffer, 1, n, fp ) );
 }
 
 static int
 vips_foreign_load_gif_open( VipsForeignLoadGif *gif, const char *filename )
 {
 	g_assert( !gif->file ); 
+	g_assert( !gif->fp ); 
+
+	if( !(gif->fp = vips__file_open_read( filename, NULL, FALSE )) ) 
+		return( -1 ); 
 
 #ifdef HAVE_GIFLIB_5
 {
 	int error; 
 
-	if( !(gif->file = DGifOpenFileName( filename, &error )) ) {
+	if( !(gif->file = 
+		DGifOpen( gif->fp, vips_giflib_file_read, &error )) ) {
 		vips_foreign_load_gif_error_vips( gif, error );
 		return( -1 ); 
 	}
 }
 #else 
-	if( !(gif->file = DGifOpenFileName( filename )) ) { 
+	if( !(gif->file = DGifOpen( gif->fp, vips_giflib_file_read )) ) { 
 		vips_foreign_load_gif_error_vips( gif, GifLastError() ); 
 		return( -1 ); 
 	}
@@ -302,39 +341,11 @@ vips_foreign_load_gif_is_a( const char *filename )
 	return( 0 );
 }
 
-static void
-vips_foreign_load_gif_parse( VipsForeignLoadGif *gif, 
-	VipsImage *out )
-{
-	vips_image_init_fields( out, 
-		gif->file->SWidth, gif->file->SHeight,
-		4, VIPS_FORMAT_UCHAR,
-		VIPS_CODING_NONE, VIPS_INTERPRETATION_sRGB, 1.0, 1.0 );
-
-	/* We will have the whole GIF frame in memory, so we can render any 
-	 * area.
-	 */
-        vips_image_pipelinev( out, VIPS_DEMAND_STYLE_ANY, NULL );
-
-	/* We need a line buffer to decompress to.
-	 */
-	gif->line = VIPS_ARRAY( gif, gif->file->SWidth, GifPixelType );
-}
-
-static int
-vips_foreign_load_gif_header( VipsForeignLoad *load )
-{
-	VipsForeignLoadGif *gif = (VipsForeignLoadGif *) load;
-
-	vips_foreign_load_gif_parse( gif, load->out ); 
-
-	return( 0 );
-}
-
 static void
 vips_foreign_load_gif_render_line( VipsForeignLoadGif *gif,
 	int width, VipsPel * restrict q, VipsPel * restrict p )
 {
+	VipsObjectClass *class = VIPS_OBJECT_GET_CLASS( gif );
 	ColorMapObject *map = gif->file->Image.ColorMap ?
 		gif->file->Image.ColorMap : gif->file->SColorMap;
 
@@ -343,8 +354,13 @@ vips_foreign_load_gif_render_line( VipsForeignLoadGif *gif,
 	for( x = 0; x < width; x++ ) {
 		VipsPel v = p[x];
 
-		if( v != gif->transparency &&
-			v < map->ColorCount ) {
+		if( v >= map->ColorCount ) {
+			vips_warn( class->nickname,
+				"%s", _( "pixel value out of range" ) );
+			continue;
+		}
+
+		if( v != gif->transparency ) { 
 			q[0] = map->Colors[v].Red;
 			q[1] = map->Colors[v].Green;
 			q[2] = map->Colors[v].Blue;
@@ -370,6 +386,8 @@ vips_foreign_load_gif_render( VipsForeignLoadGif *gif, VipsImage *out )
 {
 	VipsObjectClass *class = VIPS_OBJECT_GET_CLASS( gif );
 	GifFileType *file = gif->file;
+	ColorMapObject *map = file->Image.ColorMap ?
+		file->Image.ColorMap : file->SColorMap;
 
 	/* Check that the frame lies within our image.
 	 */
@@ -382,10 +400,27 @@ vips_foreign_load_gif_render( VipsForeignLoadGif *gif, VipsImage *out )
 		return( -1 ); 
 	}
 
+	/* Check if we have a non-greyscale colourmap for this frame.
+	 */
+	if( !gif->has_colour ) {
+		int i;
+
+		for( i = 0; i < map->ColorCount; i++ ) 
+			if( map->Colors[i].Red != map->Colors[i].Green ||
+				map->Colors[i].Green != map->Colors[i].Blue ) {
+				VIPS_DEBUG_MSG( "gifload: not mono\n" ); 
+				gif->has_colour = TRUE;
+				break;
+			}
+	}
+
 	if( file->Image.Interlace ) {
 		int i;
 
-		VIPS_DEBUG_MSG( "gifload: interlaced frame\n" ); 
+		VIPS_DEBUG_MSG( "gifload: interlaced frame of "
+			"%d x %d pixels at %d x %d\n",
+			file->Image.Width, file->Image.Height,
+			file->Image.Left, file->Image.Top ); 
 
 		for( i = 0; i < 4; i++ ) {
 			int y;
@@ -410,6 +445,11 @@ vips_foreign_load_gif_render( VipsForeignLoadGif *gif, VipsImage *out )
 	else {
 		int y;
 
+		VIPS_DEBUG_MSG( "gifload: non-interlaced frame of "
+			"%d x %d pixels at %d x %d\n",
+			file->Image.Width, file->Image.Height,
+			file->Image.Left, file->Image.Top ); 
+
 		for( y = 0; y < file->Image.Height; y++ ) {
 			VipsPel *q = VIPS_IMAGE_ADDR( out, 
 				file->Image.Left, file->Image.Top + y );
@@ -429,20 +469,31 @@ vips_foreign_load_gif_render( VipsForeignLoadGif *gif, VipsImage *out )
 }
 
 static int
-vips_foreign_load_gif_load( VipsForeignLoad *load )
+vips_foreign_load_gif_to_memory( VipsForeignLoadGif *gif, VipsImage *out )
 {
-	VipsObjectClass *class = VIPS_OBJECT_GET_CLASS( load );
-	VipsForeignLoadGif *gif = (VipsForeignLoadGif *) load;
+	VipsObjectClass *class = VIPS_OBJECT_GET_CLASS( gif );
 
 	int frame_n;
 	GifRecordType record;
 
-	vips_foreign_load_gif_parse( gif, load->real ); 
+	vips_image_init_fields( out, 
+		gif->file->SWidth, gif->file->SHeight,
+		4, VIPS_FORMAT_UCHAR,
+		VIPS_CODING_NONE, VIPS_INTERPRETATION_sRGB, 1.0, 1.0 );
+
+	/* We will have the whole GIF frame in memory, so we can render any 
+	 * area.
+	 */
+        vips_image_pipelinev( out, VIPS_DEMAND_STYLE_ANY, NULL );
+
+	/* We need a line buffer to decompress to.
+	 */
+	gif->line = VIPS_ARRAY( gif, gif->file->SWidth, GifPixelType );
 
 	/* Turn out into a memory image which we then render the GIF frames
 	 * into.
 	 */
-	if( vips_image_write_prepare( load->real ) )
+	if( vips_image_write_prepare( out ) )
 		return( -1 );
 
 	/* Scan the GIF until we have enough to have completely rendered the
@@ -467,7 +518,7 @@ vips_foreign_load_gif_load( VipsForeignLoad *load )
 				return( -1 ); 
 			}
 
-			if( vips_foreign_load_gif_render( gif, load->real ) )
+			if( vips_foreign_load_gif_render( gif, out ) )
 				return( -1 ); 
 
 			frame_n += 1;
@@ -490,11 +541,13 @@ vips_foreign_load_gif_load( VipsForeignLoad *load )
 			if( ext_code == GRAPHICS_EXT_FUNC_CODE &&
 				extension &&
 				extension[0] == 4 && 
-				extension[1] == 1 ) {
+				(extension[1] & 0x1) ) {
 				/* Bytes are 4, 1, delay low, delay high,
-				 * transparency.
+				 * transparency. Bit 1 means transparency
+				 * is being set.
 				 */
 				gif->transparency = extension[4];
+				gif->has_transparency = TRUE;
 
 				VIPS_DEBUG_MSG( "gifload: "
 					"seen transparency %d\n", 
@@ -549,6 +602,62 @@ vips_foreign_load_gif_load( VipsForeignLoad *load )
 	return( 0 );
 }
 
+static int
+vips_foreign_load_gif_load( VipsForeignLoad *load )
+{
+	VipsForeignLoadGif *gif = (VipsForeignLoadGif *) load;
+	VipsImage **t = (VipsImage **) 
+		vips_object_local_array( VIPS_OBJECT( load ), 4 );
+
+	VipsImage *im;
+
+	/* Render to a memory image.
+	 */
+	im = t[0] = vips_image_new_memory();
+	if( vips_foreign_load_gif_to_memory( gif, im ) )
+		return( -1 );
+
+	/* Depending on what we found, transform and write to load->real.
+	 */
+	if( gif->has_colour &&
+		gif->has_transparency ) {
+		/* Nothing to do.
+		 */
+	}
+	else if( gif->has_colour ) { 
+		/* RGB.
+		 */
+		if( vips_extract_band( im, &t[1], 0,
+			"n", 3,
+			NULL ) )
+			return( -1 );
+		im = t[1];
+	}
+	else if( gif->has_transparency ) {
+		/* GA. Take BA so we have neighboring channels. 
+		 */
+		if( vips_extract_band( im, &t[1], 2,
+			"n", 2,
+			NULL ) )
+			return( -1 );
+		im = t[1];
+		im->Type = VIPS_INTERPRETATION_B_W;
+	}
+	else {
+		/* G.
+		 */
+		if( vips_extract_band( im, &t[1], 0, NULL ) )
+			return( -1 );
+		im = t[1];
+		im->Type = VIPS_INTERPRETATION_B_W;
+	}
+
+	if( vips_image_write( im, load->out ) )
+		return( -1 );
+
+	return( 0 );
+}
+
 static void
 vips_foreign_load_gif_class_init( VipsForeignLoadGifClass *class )
 {
@@ -566,7 +675,6 @@ vips_foreign_load_gif_class_init( VipsForeignLoadGifClass *class )
 	load_class->get_flags_filename = 
 		vips_foreign_load_gif_get_flags_filename;
 	load_class->get_flags = vips_foreign_load_gif_get_flags;
-	load_class->load = vips_foreign_load_gif_load;
 
 	VIPS_ARG_INT( class, "page", 10,
 		_( "Page" ),
@@ -580,6 +688,7 @@ vips_foreign_load_gif_class_init( VipsForeignLoadGifClass *class )
 static void
 vips_foreign_load_gif_init( VipsForeignLoadGif *gif )
 {
+	gif->transparency = -1;
 }
 
 typedef struct _VipsForeignLoadGifFile {
@@ -605,7 +714,9 @@ vips_foreign_load_gif_file_header( VipsForeignLoad *load )
 	if( vips_foreign_load_gif_open( gif, file->filename ) ) 
 		return( -1 ); 
 
-	return( vips_foreign_load_gif_header( load ) );
+	VIPS_SETSTR( load->out->filename, file->filename );
+
+	return( vips_foreign_load_gif_load( load ) );
 }
 
 static const char *vips_foreign_gif_suffs[] = {
@@ -671,12 +782,11 @@ G_DEFINE_TYPE( VipsForeignLoadGifBuffer, vips_foreign_load_gif_buffer,
  * Read up to len bytes into buffer, return number of bytes read, 0 for EOF.
  */
 static int
-vips_foreign_load_gif_buffer_read( GifFileType *file, 
-	GifByteType *buf, int len )
+vips_giflib_buffer_read( GifFileType *file, GifByteType *buf, int n )
 {
-	VipsForeignLoadGifBuffer *buffer = (VipsForeignLoadGifBuffer *)
-		file->UserData;
-	size_t will_read = VIPS_MIN( len, buffer->bytes_to_go );
+	VipsForeignLoadGifBuffer *buffer = 
+		(VipsForeignLoadGifBuffer *) file->UserData;
+	size_t will_read = VIPS_MIN( n, buffer->bytes_to_go );
 
 	memcpy( buf, buffer->p, will_read );
 	buffer->p += will_read;
@@ -696,11 +806,10 @@ vips_foreign_load_gif_buffer_header( VipsForeignLoad *load )
 	buffer->p = buffer->buf->data;
 	buffer->bytes_to_go = buffer->buf->length;
 
-	if( vips_foreign_load_gif_open_buffer( gif, 
-		vips_foreign_load_gif_buffer_read ) ) 
+	if( vips_foreign_load_gif_open_buffer( gif, vips_giflib_buffer_read ) ) 
 		return( -1 ); 
 
-	return( vips_foreign_load_gif_header( load ) );
+	return( vips_foreign_load_gif_load( load ) );
 }
 
 static void
@@ -734,5 +843,80 @@ vips_foreign_load_gif_buffer_init( VipsForeignLoadGifBuffer *buffer )
 {
 }
 
-#endif /*HAVE_RSVG*/
+#endif /*HAVE_GIFLIB*/
+
+/**
+ * vips_gifload:
+ * @filename: file to load
+ * @out: output image
+ * @...: %NULL-terminated list of optional named arguments
+ *
+ * Optional arguments:
+ *
+ * * @page: %gint, page (frame) to read
+ *
+ * Read a GIF file into a VIPS image.  Rendering uses the giflib library.
+ *
+ * Use @page to set page number (frame number) to read.
+ *
+ * The whole GIF is rendered into memory on header access. The output image
+ * will be 1, 2, 3 or 4 bands depending on what the reader finds in the file. 
+ *
+ * See also: vips_image_new_from_file().
+ *
+ * Returns: 0 on success, -1 on error.
+ */
+int
+vips_gifload( const char *filename, VipsImage **out, ... )
+{
+	va_list ap;
+	int result;
+
+	va_start( ap, out );
+	result = vips_call_split( "gifload", ap, filename, out );
+	va_end( ap );
+
+	return( result );
+}
+
+/**
+ * vips_gifload_buffer:
+ * @buf: memory area to load
+ * @len: size of memory area
+ * @out: image to write
+ * @...: %NULL-terminated list of optional named arguments
+ *
+ * Optional arguments:
+ *
+ * * @page: %gint, page (frame) to read
+ *
+ * Read a GIF-formatted memory block into a VIPS image. Exactly as
+ * vips_gifload(), but read from a memory buffer. 
+ *
+ * You must not free the buffer while @out is active. The 
+ * #VipsObject::postclose signal on @out is a good place to free. 
+ *
+ * See also: vips_gifload().
+ *
+ * Returns: 0 on success, -1 on error.
+ */
+int
+vips_gifload_buffer( void *buf, size_t len, VipsImage **out, ... )
+{
+	va_list ap;
+	VipsBlob *blob;
+	int result;
+
+	/* We don't take a copy of the data or free it.
+	 */
+	blob = vips_blob_new( NULL, buf, len );
+
+	va_start( ap, out );
+	result = vips_call_split( "gifload_buffer", ap, blob, out );
+	va_end( ap );
+
+	vips_area_unref( VIPS_AREA( blob ) );
+
+	return( result );
+}
 
