@@ -56,6 +56,7 @@
 #define DEBUG_VERBOSE
 #define DEBUG_CREATE
 #define DEBUG
+#define DEBUG_GLOBAL
  */
 
 #ifdef HAVE_CONFIG_H
@@ -84,8 +85,22 @@ static GSList *vips__buffer_cache_all = NULL;
  */
 static const int buffer_cache_max_reserve = 2; 
 
+/* Workers have a buffer_thread in a GPrivate they have exclusive access to.
+ */
 static GPrivate *buffer_thread_key = NULL;
 
+/* All non-worker threads share a single global set of buffers protected by a
+ * mutex.
+ */
+static VipsBufferThread *vips_buffer_thread_global = NULL;
+
+#ifdef DEBUG_GLOBAL
+/* Count main thread buffers in and out.
+ */
+static int vips_buffer_thread_global_n = 0;
+static int vips_buffer_thread_global_highwater = 0;
+#endif /*DEBUG_GLOBAL*/
+
 #ifdef DEBUG
 static void *
 vips_buffer_dump( VipsBuffer *buffer, size_t *reserve, size_t *alive )
@@ -154,6 +169,12 @@ vips_buffer_dump_all( void )
 	}
 #endif /*DEBUG_CREATE*/
 #endif /*DEBUG*/
+
+#ifdef DEBUG_GLOBAL
+	printf( "buffers: %d global buffers\n", vips_buffer_thread_global_n );
+	printf( "buffers: %d high water global buffers\n", 
+		vips_buffer_thread_global_highwater ); 
+#endif /*DEBUG_GLOBAL*/
 }
 
 static void
@@ -176,6 +197,8 @@ vips_buffer_free( VipsBuffer *buffer )
 static void
 buffer_thread_free( VipsBufferThread *buffer_thread )
 {
+	/* We only come here from workers, so no need to lock.
+	 */
 	VIPS_FREEF( g_hash_table_destroy, buffer_thread->hash );
 	VIPS_FREE( buffer_thread );
 }
@@ -184,8 +207,8 @@ buffer_thread_free( VipsBufferThread *buffer_thread )
  *
  * - on thread shutdown, the enclosing hash is destroyed, and that will 
  *   trigger this via GDestroyNotify.
- * - if the image is closed, buffer_cache_image_postclose() fires and will call
- *   this
+ * - if the BufferCache has been allocated by the main thread,  this will be
+ *   triggered from postclose on the image
  *
  * These can happen in either order. 
  */
@@ -223,35 +246,39 @@ buffer_cache_free( VipsBufferCache *cache )
 	}
 	VIPS_FREEF( g_slist_free, cache->reserve );
 
-	cache->proxy->cache = NULL;
-	cache->proxy = NULL;
-
 	g_free( cache );
 }
 
 static void
-buffer_cache_proxy_image_postclose( VipsImage *im, VipsBufferCacheProxy *proxy )
+buffer_cache_image_postclose( VipsImage *im, VipsBufferCache *cache )
 {
-	g_assert( proxy->im == im );
+	VipsBufferThread *buffer_thread = cache->buffer_thread;
 
-	if( proxy->cache ) {
-		VipsBufferCache *cache = proxy->cache;
-		VipsBufferThread *buffer_thread = cache->buffer_thread;
-		VipsImage *im = cache->im;
+	/* Runs to clean up main thread buffers on image close.
+	 */
+	g_assert( cache->im == im );
+	g_assert( !vips_thread_isworker() );
 
-		g_assert( cache->im == im );
+	/* All non-worker threads come through here, so we need to lock around
+	 * changes to the global buffer_thread.
+	 */
+	g_mutex_lock( vips__global_lock );
 
-		g_hash_table_remove( buffer_thread->hash, im );
-	}
+	g_hash_table_remove( buffer_thread->hash, im );
 
-	g_free( proxy ); 
+#ifdef DEBUG_GLOBAL
+	vips_buffer_thread_global_n -= 1;
+	printf( "buffer_cache_image_postclose: %d global buffers\n",
+		vips_buffer_thread_global_n ); 
+#endif /*DEBUG_GLOBAL*/
+
+	g_mutex_unlock( vips__global_lock );
 }
 
 static VipsBufferCache *
 buffer_cache_new( VipsBufferThread *buffer_thread, VipsImage *im )
 {
 	VipsBufferCache *cache;
-	VipsBufferCacheProxy *proxy;
 
 	cache = g_new( VipsBufferCache, 1 );
 	cache->buffers = NULL;
@@ -261,17 +288,28 @@ buffer_cache_new( VipsBufferThread *buffer_thread, VipsImage *im )
 	cache->reserve = NULL;
 	cache->n_reserve = 0;
 
-	proxy = g_new( VipsBufferCacheProxy, 1 );
-	proxy->im = im;
-	proxy->cache = cache;
-	cache->proxy = proxy;
-	
-	/* Free memory on image close. We can't do this on thread exit since
-	 * some buffers will be made from the main thread and that won't exit
-	 * until program termination.
+	/* VipsBufferCache allocated from worker threads will be freed when
+	 * workers shut down. This won't happen for VipsBufferCache allocated 
+	 * from the main thread, since (obviously) thread shutdown will never 
+	 * happen. In this case, we need to free resources on image close.
 	 */
-	g_signal_connect( im, "postclose", 
-		G_CALLBACK( buffer_cache_proxy_image_postclose ), proxy );
+	if( !vips_thread_isworker() ) {
+		g_signal_connect( im, "postclose", 
+			G_CALLBACK( buffer_cache_image_postclose ), cache );
+
+#ifdef DEBUG_GLOBAL
+		/* No need to lock. Main thread buffer_cache_new() calls are
+		 * always inside a lock already.
+		 */
+		vips_buffer_thread_global_n += 1;
+		vips_buffer_thread_global_highwater = VIPS_MAX( 
+			vips_buffer_thread_global_highwater, 
+			vips_buffer_thread_global_n ); 
+
+		printf( "buffer_cache_new: %d global buffers\n",
+			vips_buffer_thread_global_n ); 
+#endif /*DEBUG_GLOBAL*/
+	}
 
 #ifdef DEBUG_CREATE
 	g_mutex_lock( vips__global_lock );
@@ -307,12 +345,32 @@ buffer_thread_get( void )
 {
 	VipsBufferThread *buffer_thread;
 
-	if( !(buffer_thread = g_private_get( buffer_thread_key )) ) {
-		buffer_thread = buffer_thread_new();
-		g_private_set( buffer_thread_key, buffer_thread );
-	}
+	if( vips_thread_isworker() ) {
+		/* Workers get a private set of buffers.
+		 */
+		if( !(buffer_thread = g_private_get( buffer_thread_key )) ) {
+			buffer_thread = buffer_thread_new();
+			g_private_set( buffer_thread_key, buffer_thread );
+		}
 
-	g_assert( buffer_thread->thread == g_thread_self() ); 
+		g_assert( buffer_thread->thread == g_thread_self() ); 
+	}
+	else {
+		/* All main threads share a single set of buffers. 
+		 */
+		g_mutex_lock( vips__global_lock );
+
+		if( !vips_buffer_thread_global ) {
+			vips_buffer_thread_global = buffer_thread_new(); 
+
+			/* Shared by many threads, so no checking.
+			 */
+			vips_buffer_thread_global->thread = NULL;
+		}
+		buffer_thread = vips_buffer_thread_global;
+
+		g_mutex_unlock( vips__global_lock );
+	}
 
 	return( buffer_thread );
 }
@@ -324,13 +382,20 @@ buffer_cache_get( VipsImage *im )
 
 	VipsBufferCache *cache;
 
+	if( !vips_thread_isworker() ) 
+		g_mutex_lock( vips__global_lock );
+
 	if( !(cache = (VipsBufferCache *) 
 		g_hash_table_lookup( buffer_thread->hash, im )) ) {
 		cache = buffer_cache_new( buffer_thread, im );
 		g_hash_table_insert( buffer_thread->hash, im, cache );
 	}
 
-	g_assert( cache->thread == g_thread_self() ); 
+	if( !vips_thread_isworker() ) 
+		g_mutex_unlock( vips__global_lock );
+
+	g_assert( !cache->thread ||
+		cache->thread == g_thread_self() ); 
 
 	return( cache ); 
 }
@@ -374,8 +439,10 @@ vips_buffer_undone( VipsBuffer *buffer )
 			g_thread_self(), buffer, cache );
 #endif /*DEBUG_VERBOSE*/
 
-		g_assert( cache->thread == g_thread_self() );
-		g_assert( cache->buffer_thread->thread == cache->thread );
+		g_assert( !cache->thread ||
+			cache->thread == g_thread_self() );
+		g_assert( !cache->thread ||
+			cache->buffer_thread->thread == cache->thread );
 		g_assert( g_slist_find( cache->buffers, buffer ) );
 		g_assert( cache->buffer_thread == buffer_thread_get() );
 
@@ -410,8 +477,13 @@ vips_buffer_unref( VipsBuffer *buffer )
 	buffer->ref_count -= 1;
 
 	if( buffer->ref_count == 0 ) {
-		VipsImage *im = buffer->im;
-		VipsBufferCache *cache = buffer_cache_get( im ); 
+		/* We are not always the creating thread, for example if we 
+		 * come here during vips_region_dispose(). cache may have been
+		 * NULLed out during thread exit. 
+		VipsBufferCache *cache = buffer->cache;
+		 */
+
+	VipsBufferCache *cache = buffer_cache_get( buffer->im ); 
 
 #ifdef DEBUG_VERBOSE
 		if( !buffer->done )
@@ -422,7 +494,8 @@ vips_buffer_unref( VipsBuffer *buffer )
 
 		/* Place on this thread's reserve list for reuse.
 		 */
-		if( cache->n_reserve < buffer_cache_max_reserve ) { 
+		if( cache &&
+			cache->n_reserve < buffer_cache_max_reserve ) { 
 			g_assert( !buffer->cache ); 
 
 			cache->reserve = 
@@ -633,7 +706,8 @@ buffer_thread_destroy_notify( VipsBufferThread *buffer_thread )
 	/* We only come here if vips_thread_shutdown() was not called for this
 	 * thread. Do our best to clean up.
 	 *
-	 * GPrivate has stopped working, be careful not to touch that. 
+	 * GPrivate has stopped working by this point in destruction, be 
+	 * careful not to touch that. 
 	 */
 	buffer_thread_free( buffer_thread );
 }
